#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## VII Coupler
# Paul Dechamps

from dart.viscous.Drivers.DGroupConstructor import GroupConstructor

from fwk.coloring import ccolors
import math

import fwk
import dart.pyVII.vUtils as viscU
import numpy as np

class Coupler:
  def __init__(self, _iSolverAPI, vSolver) -> None:
    self.iSolverAPI = _iSolverAPI
    self.vSolver = vSolver

    self.maxCouplIter = 150
    self.couplTol = 1e-4

    self.tms=fwk.Timers()
    self.resetInv = False
    pass

  def Run(self):

    # Initilize meshes in c++ objects

    self.nElmUpperPrev = 0
    couplIter = 0
    cdPrev = 0
    print('- AoA: {:<2.2f} Mach: {:<1.1f} Re: {:<2.1f}e6'
    .format(self.iSolverAPI.GetAlpha()*180/math.pi, self.iSolverAPI.GetMinf(), self.vSolver.GetRe()/1e6))
    print('{:>5s}| {:>10s} {:>10s} | {:>8s} {:>6s} {:>8s}'.format('It', 'Cl', 'Cd', 'xtrT', 'xtrB', 'Error'))
    print(' --------------------------------------------------------')
    while couplIter < self.maxCouplIter:

      # Run inviscid solver
      self.tms['inviscid'].start()
      if self.resetInv:
        self.iSolverAPI.ResetSolver()
      self.iSolverAPI.RunSolver()
      self.tms['inviscid'].stop()

      # Extract inviscid boundary
      self.tms['processing'].start()
      self.iSolverAPI.GetInvBC(self.vSolver, couplIter)
      self.tms['processing'].stop()

      # Run viscous solver
      self.tms['viscous'].start()
      exitCode = self.vSolver.Run(couplIter)
      self.tms['viscous'].stop()
      # print('Viscous ops terminated with exit code ', exitCode)

      self.tms['processing'].start()
      self.iSolverAPI.SetBlowingVel(self.vSolver)
      self.tms['processing'].stop()

      error = abs((self.vSolver.Cdt - cdPrev)/self.vSolver.Cdt) if self.vSolver.Cdt != 0 else 1
      
      if error  <= self.couplTol and exitCode==0:

        print(ccolors.ANSI_GREEN,'{:>4.0f}| {:>10.5f} {:>10.5f} | {:>8.2f} {:>6.2f} {:>8.2f}\n'.format(couplIter,
        self.iSolverAPI.GetCl(), self.vSolver.Cdt, self.vSolver.Getxtr(0,0), self.vSolver.Getxtr(0,1), np.log10(error)),ccolors.ANSI_RESET)
        return 0
      
      cdPrev = self.vSolver.Cdt


      if couplIter%5==0:
        print(' {:>4.0f}| {:>10.5f} {:>10.5f} | {:>8.2f} {:>6.2f} {:>8.2f}'.format(couplIter,
        self.iSolverAPI.GetCl(), self.vSolver.Cdt, self.vSolver.Getxtr(0,0), self.vSolver.Getxtr(0,1), np.log10(error)))

      couplIter += 1
      """if couplIter == 5:
        quit()"""
      self.tms['processing'].stop()
    else:
      print(ccolors.ANSI_RED, 'Maximum number of {:<.0f} coupling iterations reached'.format(self.maxCouplIter), ccolors.ANSI_RESET)
      return couplIter


  def RunPolar(self, alphaSeq):
    
    import math

    self.iSolver.printOn = False
    self.vSolver.printOn = False

    coeffTable = np.empty((3,len(alphaSeq)))
    coeffTable[0,:] = alphaSeq

    for ind, alpha in enumerate(alphaSeq):
      self.iSolver.pbl.update(alpha*math.pi/180)

      self.Run()

      # Collect Coeff

      coeffTable[1,ind] = self.iSolver.Cl
      coeffTable[2,ind] = self.vSolver.Cdt

    return coeffTable

  def __ImposeBlwVel(self):

    # Retreive and set blowing velocities.

    for iRegion in range(3):
      self.blwVel[iRegion] = np.asarray(self.vSolver.ExtractBlowingVel(iRegion))
      self.deltaStar[iRegion] = np.asarray(self.vSolver.ExtractDeltaStar(iRegion))
      self.xx[iRegion] = np.asarray(self.vSolver.ExtractXx(iRegion))

    blwVelAF = np.concatenate((self.blwVel[0], self.blwVel[1]))
    deltaStarAF = np.concatenate((self.deltaStar[0], self.deltaStar[1][1:])) # Remove stagnation point doublon.
    xxAF = np.concatenate((self.xx[0], self.xx[1][1:])) # Remove stagnation point doublon.
    blwVelWK = self.blwVel[2]
    deltaStarWK = self.deltaStar[2]
    xxWK = self.xx[2]

    self.group[0].u = blwVelAF[self.group[0].connectListElems.argsort()]
    self.group[1].u = blwVelWK[self.group[1].connectListElems.argsort()]

    self.group[0].deltaStar = deltaStarAF[self.group[0].connectListNodes.argsort()]
    self.group[1].deltaStar = deltaStarWK[self.group[1].connectListNodes.argsort()]

    self.group[0].xx = xxAF[self.group[0].connectListNodes.argsort()]
    self.group[1].xx = xxWK[self.group[1].connectListNodes.argsort()]

    for n in range(0, len(self.group)):
      for i in range(0, self.group[n].nE):
        self.group[n].boundary.setU(i, self.group[n].u[i])

  def __ImposeInvBC(self, couplIter):

    for n in range(0, len(self.group)):

      for i in range(0, len(self.group[n].boundary.nodes)):

        self.group[n].v[i,0] = self.iSolver.U[self.group[n].boundary.nodes[i].row][0]
        self.group[n].v[i,1] = self.iSolver.U[self.group[n].boundary.nodes[i].row][1]
        self.group[n].v[i,2] = self.iSolver.U[self.group[n].boundary.nodes[i].row][2]
        self.group[n].Me[i] = self.iSolver.M[self.group[n].boundary.nodes[i].row]
        self.group[n].rhoe[i] = self.iSolver.rho[self.group[n].boundary.nodes[i].row]
    
    dataIn=[None,None]
    for n in range(0,len(self.group)):
      self.group[n].connectListNodes, self.group[n].connectListElems, dataIn[n]  = self.group[n].connectList()
    fMeshDict, _, _ = GroupConstructor().mergeVectors(dataIn, 0, 1)

    if ((len(fMeshDict['locCoord'][:fMeshDict['bNodes'][1]]) != self.nElmUpperPrev) or couplIter==0):

      # print('STAG POINT MOVED')
      
      # Initialize mesh
      self.vSolver.InitMeshes(0, fMeshDict['locCoord'][:fMeshDict['bNodes'][1]], fMeshDict['globCoord'][:fMeshDict['bNodes'][1]])
      self.vSolver.InitMeshes(1, fMeshDict['locCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
      self.vSolver.InitMeshes(2, fMeshDict['locCoord'][fMeshDict['bNodes'][2]:], fMeshDict['globCoord'][fMeshDict['bNodes'][2]:])

      self.vSolver.SendExtVariables(0, fMeshDict['locCoord'][:fMeshDict['bNodes'][1]], fMeshDict['deltaStarExt'][:fMeshDict['bNodes'][1]])
      self.vSolver.SendExtVariables(1, fMeshDict['locCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['deltaStarExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])    
      self.vSolver.SendExtVariables(2, fMeshDict['locCoord'][fMeshDict['bNodes'][2]:], fMeshDict['deltaStarExt'][fMeshDict['bNodes'][2]:])
    
    else:

      self.vSolver.SendExtVariables(0, fMeshDict['xxExt'][:fMeshDict['bNodes'][1]], fMeshDict['deltaStarExt'][:fMeshDict['bNodes'][1]])
      self.vSolver.SendExtVariables(1, fMeshDict['xxExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['deltaStarExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
      self.vSolver.SendExtVariables(2, fMeshDict['xxExt'][fMeshDict['bNodes'][2]:], fMeshDict['deltaStarExt'][fMeshDict['bNodes'][2]:])

    self.nElmUpperPrev = len(fMeshDict['locCoord'][:fMeshDict['bNodes'][1]])
    self.vSolver.SendInvBC(0, fMeshDict['vtExt'][:fMeshDict['bNodes'][1]], fMeshDict['Me'][:fMeshDict['bNodes'][1]], fMeshDict['rho'][:fMeshDict['bNodes'][1]])
    self.vSolver.SendInvBC(1, fMeshDict['vtExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['Me'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['rho'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
    self.vSolver.SendInvBC(2, fMeshDict['vtExt'][fMeshDict['bNodes'][2]:], fMeshDict['Me'][fMeshDict['bNodes'][2]:], fMeshDict['rho'][fMeshDict['bNodes'][2]:])
    # print('Inviscid boundary imposed')
  


  



