/* ----------------------------------------------------------------------------------- */
/*                                                                                     */
/*                       Coupled integral boundary layer solver                        */
/*                                                                                     */
/*  Université de Liège                                                                */
/*  February 2022                                                                      */
/*  Author: Paul Dechamps                                                              */
/*  v0.1                                                                               */
/* ----------------------------------------------------------------------------------- */

/* --- Project includes -------------------------------------------------------------- */

#include "wViscSolver.h"
#include "wBLRegion.h"
#include "wTimeSolver.h"
#include "wInterpolator.h"
#include <iostream>
#include <vector>
#include <cmath>

/* --- Namespaces -------------------------------------------------------------------- */

using namespace dart;

/* ----------------------------------------------------------------------------------- */

/**
 * @brief Viscous solver and boundary layer(s) initialization
 */
ViscSolver::ViscSolver(double _Re, double _Minf, double _CFL0, unsigned int nSections)
{

    PrintBanner();

    /* Freestream parameters */

    Re = _Re;     /* Freestream Reynolds number */

    CFL0 = _CFL0;

    /* Initialzing boundary layer */
    Sections.resize(nSections);
    for (size_t iSec=0; iSec < nSections; ++iSec)
    {
        for (size_t i = 0; i < 3; ++i)
        {
            BLRegion *region = new BLRegion();
            Sections[iSec].push_back(region);
        }
        Sections[iSec][0]->name = "upper";
        Sections[iSec][1]->name = "lower";
        Sections[iSec][2]->name = "wake";
    }

    stagPtMvmt.resize(Sections.size(), false);

    /* Temporal solver initialization */

    tSolver = new TimeSolver(_CFL0, _Minf, Re);

    /* User input numerical parameters */

    std::cout << "--- Viscous solver setup ---\n"
              << "Reynolds number: " << Re << " \n"
              << "Freestream Mach number: " << _Minf << " \n"
              << "Initial CFL number: " << CFL0 << " \n"
              << std::endl;

    std::cout << "Boundary layer initialized" << std::endl;

} // end ViscSolver

/**
 * @brief Viscous solver and subsequent dependancies destruction
 */
ViscSolver::~ViscSolver()
{   
    for (size_t iSec=0; iSec<Sections.size(); ++iSec)
    {
        for (size_t i = 0; i < Sections[iSec].size(); ++i)
        {
            delete Sections[iSec][i];
        }
    }

    delete tSolver;
    PrintBanner();

    std::cout << "~ViscSolver()\n";
} // end ~ViscSolver

/**
 * @brief Runs viscous operations. Temporal solver parametrisation is first adapted,
 * Solutions are initialized than time marched towards steady state successively for each points.
 */
int ViscSolver::Run(unsigned int couplIter)
{   
    bool lockTrans;
    int solverExitCode = 0;
    int pointExitCode;

    for (size_t iSec=0; iSec<Sections.size(); ++iSec)
    {

        /* Prepare time integration */

        bool resetSolution = false;
        if ((couplIter == 0) ||
             Sections[iSec][0]->mesh->GetDiffnElm()!=0 ||
             Sections[iSec][1]->mesh->GetDiffnElm()!=0 ||
             Sections[iSec][2]->mesh->GetDiffnElm()!=0 ||
            (stagPtMvmt[iSec]))
        {
            resetSolution = true;
        }
        else
        {
            resetSolution = false;
        }
        /*std::cout << "resetSolution " << resetSolution << std::endl;

        std::cout << Sections[iSec][0]->mesh->GetDiffnElm()  <<
             Sections[iSec][1]->mesh->GetDiffnElm()  <<
             Sections[iSec][2]->mesh->GetDiffnElm() << std::endl;*/

        if (resetSolution)
        {
            tSolver->SetCFL0(1);
            tSolver->SetitFrozenJac(1);
            tSolver->SetinitSln(true);

            /* Special case where the external flow mesh cannot be used for the interaction law */
            for (size_t iRegion=0; iRegion<Sections[iSec].size(); ++iRegion)
            {
                std::vector<double> xxExtCurr;
                for (size_t iPoint=0; iPoint<Sections[iSec][iRegion]->mesh->GetnMarkers(); ++iPoint)
                {
                    xxExtCurr.push_back(Sections[iSec][iRegion]->mesh->GetLoc(iPoint));
                }
                Sections[iSec][iRegion]->mesh->SetExt(xxExtCurr);
            }

            if (couplIter != 0 && (Sections[iSec][0]->mesh->GetDiffnElm()))
            {
                stagPtMvmt[iSec] = true;
                if (printOn)
                {
                    std::cout << "Stagnation point moved" << std::endl;
                }
            }
            else
            {
                stagPtMvmt[iSec] = false;
            }
            if (printOn)
            {
                std::cout << "Restart solution" << std::endl;
            }
        }

        else
        {
            tSolver->SetCFL0(CFL0);
            tSolver->SetitFrozenJac(5);
            tSolver->SetinitSln(false);
            stagPtMvmt[iSec] = false;
            if (printOn)
            {
                std::cout << "Updating solution" << std::endl;
            }
        }

        for (size_t iRegion=0; iRegion<Sections[iSec].size(); ++iRegion)
        { 
            Sections[iSec][iRegion]->mesh->prevnMarkers = Sections[iSec][iRegion]->mesh->GetnMarkers();
        }

        /* Set boundary condition */

        Sections[iSec][0]->xtr = 1.0; /* Upper side initial xtr */
        Sections[iSec][1]->xtr = 1.0; /* Lower side initial xtr */
        Sections[iSec][2]->xtr = 0.0; /* Wake initial xtr (full turbulent) */

        Sections[iSec][0]->SetStagBC(Re);
        Sections[iSec][1]->SetStagBC(Re);


        /* Loop over the regions */
        for (size_t iRegion = 0; iRegion < Sections[iSec].size(); ++iRegion)
        {
            if (printOn)
            {
                std::cout << "- " << Sections[iSec][iRegion]->name << " region";
            }
            lockTrans = false;
            
            /* Reset regime */
            if (iRegion < 2)
            {
                for (size_t k = 0; k < Sections[iSec][iRegion]->mesh->GetnMarkers(); k++)
                {
                    Sections[iSec][iRegion]->Regime[k] = 0;
                }
            }
            else
            {
                for (size_t k = 0; k < Sections[iSec][iRegion]->mesh->GetnMarkers(); k++)
                {
                    Sections[iSec][iRegion]->Regime[k] = 1;
                }

                /* Impose wake boundary condition */
                std::vector<double> UpperCond;
                std::vector<double> LowerCond;
                UpperCond = std::vector<double>(Sections[iSec][0]->U.end() - Sections[iSec][0]->GetnVar(), Sections[iSec][0]->U.end()); /* Base std::vector constructor to slice */
                LowerCond = std::vector<double>(Sections[iSec][1]->U.end() - Sections[iSec][1]->GetnVar(), Sections[iSec][1]->U.end()); /* Base std::vector constructor to slice */
                Sections[iSec][iRegion]->SetWakeBC(Re, UpperCond, LowerCond);
                lockTrans = true;
            }

            /* Loop over points */
            for (size_t iPoint = 1; iPoint < Sections[iSec][iRegion]->mesh->GetnMarkers(); ++iPoint)
            {

                /* Initialize solution at point */

                tSolver->InitialCondition(iPoint, Sections[iSec][iRegion]);

                /* Time integration */
                /*std::cout << " iSec " << iSec
                          << " iPoint " << iPoint
                          << " Loc " << Sections[iSec][iRegion]->mesh->GetLoc(iPoint)
                          << " Vt " << Sections[iSec][iRegion]->invBnd->GetVt(iPoint)
                          << " Me " << Sections[iSec][iRegion]->invBnd->GetMe(iPoint)
                          << " xxExt " << Sections[iSec][iRegion]->mesh->GetExt(iPoint)
                          << " DeltaStarExt " << Sections[iSec][iRegion]->invBnd->GetDeltaStar(iPoint)
                          << std::endl;
                Sections[iSec][iRegion]->PrintSolution(iPoint);*/

                pointExitCode = tSolver->Integration(iPoint, Sections[iSec][iRegion]);

                if (pointExitCode != 0)
                {
                    if (printOn)
                    {
                        std::cout << "Point " << iPoint << ": Convergence terminated with exitcode " << pointExitCode << std::endl;
                    }
                    solverExitCode = -1; // Convergence failed
                }

                /* Transition */

                if (!lockTrans)
                {
                    CheckWaves(iPoint, Sections[iSec][iRegion]);

                    // Free transition.
                    if (Sections[iSec][iRegion]->U[iPoint * Sections[iSec][iRegion]->GetnVar() + 2] >= Sections[iSec][iRegion]->GetnCrit())
                    {
                        AverageTransition(iPoint, Sections[iSec][iRegion], 0);
                        if (printOn)
                        {
                            std::cout << ", free transition x/c = "
                                      << Sections[iSec][iRegion]->xtr
                                      << ". Marker " << Sections[iSec][iRegion]->transMarker
                                      << std::endl;
                        }
                        lockTrans = true;
                    } // End if
                }     // End if (!lockTrans).

            } // End for iPoints
            Sections[iSec][iRegion]->ComputeBlwVel();
        }     // End for iRegion

    } // End for iSections

    ComputeDrag(Sections[0]);
    if (printOn)
    {
        std::cout << "\n" << std::endl;
    }

    return solverExitCode; // exit code
}

/**
 * @brief Check whether instabilities are growing in the laminar boundary layer.
 */
void ViscSolver::CheckWaves(size_t iPoint, BLRegion *bl)
{
    /* Determine if the amplification waves start growing or not. */

    double logRet_crit = 2.492 * pow(1 / (bl->Hk[iPoint] - 1), 0.43) + 0.7 * (tanh(14 * (1 / (bl->Hk[iPoint] - 1)) - 9.24) + 1.0);

    if (bl->Ret[iPoint] > 0)
    { // Avoid log of negative number.
        if (log10(bl->Ret[iPoint]) <= logRet_crit - 0.08)
        {
            bl->U[iPoint * bl->GetnVar() + 2] = 0; // Set N to 0 at that point if waves do not grow.
        }
    }
}

/**
 * @brief Computes turbulent solution @ the transition point and averages solutions.
 */
void ViscSolver::AverageTransition(size_t iPoint, BLRegion *bl, int forced)
{
    /* Averages solution on transition marker */

    unsigned int nVar = bl->GetnVar();

    // Save laminar solution.
    std::vector<double> lamSol(nVar);
    for (size_t k = 0; k < lamSol.size(); ++k)
    {
        lamSol[k] = bl->U[iPoint * nVar + k];
    }

    // Set up turbulent portion boundary condition.
    bl->transMarker = iPoint; /* Save transition marker */

    /* Loop over remaining points in the region */
    for (size_t k = 0; k < bl->mesh->GetnMarkers() - iPoint; ++k)
    {
        bl->Regime[iPoint + k] = 1; // Set Turbulent regime.
    }

    /* Compute transition location */
    bl->xtr = (bl->GetnCrit() - bl->U[(iPoint - 1) * nVar + 2]) * ((bl->mesh->Getx(iPoint) - bl->mesh->Getx(iPoint - 1)) / (bl->U[iPoint * nVar + 2] - bl->U[(iPoint - 1) * nVar + 2])) + bl->mesh->Getx(iPoint - 1);

    /*  Percentage of the subsection corresponding to a turbulent flow */
    double avTurb = (bl->mesh->Getx(iPoint) - bl->xtr) / (bl->mesh->Getx(iPoint) - bl->mesh->Getx(iPoint - 1));
    double avLam = 1 - avTurb;

    /* Impose boundary condition */
    std::vector<double> turbParam1 = bl->closSolver->TurbulentClosures(bl->U[(iPoint - 1) * nVar + 0], bl->U[(iPoint - 1) * nVar + 1], 0.03, bl->Ret[iPoint - 1], bl->invBnd->GetMe(iPoint - 1), bl->name);
    bl->Cteq[iPoint - 1] = turbParam1[6];
    bl->U[(iPoint - 1) * nVar + 4] = 0.7 * bl->Cteq[iPoint - 1];

    /* Avoid starting with ill conditioned IC for turbulent computation. (Regime was switched line 183) */
    /* These initial guess values do not influence the converged solution */
    bl->U[iPoint * nVar + 4] = bl->U[(iPoint - 1) * nVar + 4]; // IC of transition point = turbulent BC (imposed @ previous point).
    bl->U[iPoint * nVar + 1] = 1.515;                          // Because we expect a significant drop of the shape factor.

    /* Solve point in turbulent configuration */
    int exitCode = tSolver->Integration(iPoint, bl);

    if (exitCode != 0)
    {
        std::cout << "Warning level 3: Transition point turbulent computation terminated with exit code " << exitCode << std::endl;
    }

    /* Average both solutions (Now solution @ iPoint is the turbulent solution) */
    bl->U[iPoint * nVar + 0] = avLam * lamSol[0] + avTurb * bl->U[(iPoint)*nVar + 0]; /* Theta */
    bl->U[iPoint * nVar + 1] = avLam * lamSol[1] + avTurb * bl->U[(iPoint)*nVar + 1]; /* H */
    bl->U[iPoint * nVar + 2] = 9;                                                     /* N */
    bl->U[iPoint * nVar + 3] = avLam * lamSol[3] + avTurb * bl->U[(iPoint)*nVar + 3]; /* ue */
    bl->U[iPoint * nVar + 4] = avTurb * bl->U[(iPoint)*nVar + 4];                     /* Ct */

    /* Recompute closures. (The turbulent BC @ iPoint - 1 does not influence laminar closure relations). */
    std::vector<double> lamParam = bl->closSolver->LaminarClosures(bl->U[(iPoint - 1) * nVar + 0], bl->U[(iPoint - 1) * nVar + 1], bl->Ret[iPoint - 1], bl->invBnd->GetMe(iPoint - 1), bl->name);
    bl->Hstar[iPoint - 1] = lamParam[0];
    bl->Hstar2[iPoint - 1] = lamParam[1];
    bl->Hk[iPoint - 1] = lamParam[2];
    bl->Cf[iPoint - 1] = lamParam[3];
    bl->Cd[iPoint - 1] = lamParam[4];
    bl->Delta[iPoint - 1] = lamParam[5];
    bl->Cteq[iPoint - 1] = lamParam[6];
    bl->Us[iPoint - 1] = lamParam[7];

    std::vector<double> turbParam = bl->closSolver->TurbulentClosures(bl->U[(iPoint)*nVar + 0], bl->U[(iPoint)*nVar + 1], bl->U[(iPoint)*nVar + 4], bl->Ret[iPoint], bl->invBnd->GetMe(iPoint), bl->name);
    bl->Hstar[iPoint] = turbParam[0];
    bl->Hstar2[iPoint] = turbParam[1];
    bl->Hk[iPoint] = turbParam[2];
    bl->Cf[iPoint] = turbParam[3];
    bl->Cd[iPoint] = turbParam[4];
    bl->Delta[iPoint] = turbParam[5];
    bl->Cteq[iPoint] = turbParam[6];
    bl->Us[iPoint] = turbParam[7];
}

/**
 * @brief Friction, pressure and total drag computation of the configuration.
 * 
 * Cdtot = theta * Ue^((H+5)/2).
 * Cdf = integral(Cf dx)_upper + integral(Cf dx)_lower.
 * Cdp = Cdtot - Cdf.
 */
void ViscSolver::ComputeDrag(std::vector<BLRegion *> bl)
{

    unsigned int nVar = bl[0]->GetnVar();
    unsigned int lastWkPt = (bl[2]->mesh->GetnMarkers() - 1) * nVar;

    /* Normalize friction and dissipation coefficients. */

    std::vector<double> frictioncoeff_upper;
    for (size_t k = 0; k < bl[0]->mesh->GetnMarkers(); ++k)
    {
        frictioncoeff_upper.push_back(bl[0]->U[k * nVar + 3] * bl[0]->U[k * nVar + 3] * bl[0]->Cf[k]);
    }
    std::vector<double> frictioncoeff_lower;
    for (size_t k = 0; k < bl[1]->mesh->GetnMarkers(); ++k)
    {
        frictioncoeff_lower.push_back(bl[1]->U[k * nVar + 3] * bl[1]->U[k * nVar + 3] * bl[1]->Cf[k]);
    }

    /* Compute friction drag coefficient. */

    double Cdf_upper = 0.0;
    for (size_t k = 1; k < bl[0]->mesh->GetnMarkers(); ++k)
    {
        Cdf_upper += (frictioncoeff_upper[k] + frictioncoeff_upper[k - 1]) * (bl[0]->mesh->Getx(k) - bl[0]->mesh->Getx(k - 1)) / 2;
    }
    double Cdf_lower = 0.0;
    for (size_t k = 1; k < bl[1]->mesh->GetnMarkers(); ++k)
    {
        Cdf_lower += (frictioncoeff_lower[k] + frictioncoeff_lower[k - 1]) * (bl[1]->mesh->Getx(k) - bl[1]->mesh->Getx(k - 1)) / 2;
    }

    Cdf = Cdf_upper + Cdf_lower; // No friction in the wake

    /* Compute total drag coefficient. */

    Cdt = 2 * bl[2]->U[lastWkPt + 0] * pow(bl[2]->U[lastWkPt + 3], (bl[2]->U[lastWkPt + 1] + 5) / 2);

    /* Compute pressure drag coefficient. */

    Cdp = Cdt - Cdf;
}

void ViscSolver::SetGlobMesh(size_t iSec, size_t iRegion, std::vector<double> _x, std::vector<double> _y, std::vector<double> _z)
{
    std::cout << "len x " << _x.size() << std::endl;
    Sections[iSec][iRegion]->SetMesh(_x, _y, _z);
}

void ViscSolver::SetVelocities(size_t iSec, size_t iRegion, std::vector<double> vx, std::vector<double> vy, std::vector<double> vz)
{
    std::vector<double> Vt(vx.size());
    double alpha=0;

    for (size_t iPoint=0; iPoint<vx.size()-1; ++iPoint)
    {
        alpha = std::atan2(Sections[iSec][iRegion]->mesh->Gety(iPoint+1) - Sections[iSec][iRegion]->mesh->Gety(iPoint),
                           Sections[iSec][iRegion]->mesh->Getx(iPoint+1) - Sections[iSec][iRegion]->mesh->Getx(iPoint));
        Vt[iPoint] = vx[iPoint]*std::cos(alpha) + vy[iPoint]*std::sin(alpha);
    }
    Vt.back() = vx.back()*cos(alpha) + vy.back()*std::sin(alpha);

    Sections[iSec][iRegion]->invBnd->SetVt(Vt);
}

void ViscSolver::SetMe(size_t iSec, size_t iRegion, std::vector<double> _Me)
{
    Sections[iSec][iRegion]->invBnd->SetMe(_Me);
}

void ViscSolver::SetRhoe(size_t iSec, size_t iRegion, std::vector<double> _Rhoe)
{
    Sections[iSec][iRegion]->invBnd->SetRhoe(_Rhoe);
}

void ViscSolver::SetxxExt(size_t iSec, size_t iRegion, std::vector<double> _xxExt)
{   
    std::cout << "len xxExt " << _xxExt.size() << std::endl;
    Sections[iSec][iRegion]->mesh->SetExt(_xxExt);
}

void ViscSolver::SetDeltaStarExt(size_t iSec, size_t iRegion, std::vector<double> _DeltaStarExt)
{
    std::cout << "len DeltaStarExt " << _DeltaStarExt.size() << std::endl;
    Sections[iSec][iRegion]->invBnd->SetDeltaStar(_DeltaStarExt);
}

std::vector<double> ViscSolver::ExtractBlowingVel(size_t iSec, size_t iRegion)
{
    return Sections[iSec][iRegion]->Blowing;
}
std::vector<double> ViscSolver::Extractxx(size_t iSec, size_t iRegion)
{   
    std::vector<double> xx(Sections[iSec][iRegion]->mesh->GetnMarkers(),0);
    for (size_t iPoint=0; iPoint<Sections[iSec][iRegion]->mesh->GetnMarkers(); ++iPoint)
    {
        xx[iPoint] = Sections[iSec][iRegion]->mesh->GetLoc(iPoint);
    }
    return xx;
}
std::vector<double> ViscSolver::ExtractDeltaStar(size_t iSec, size_t iRegion)
{
    return Sections[iSec][iRegion]->DeltaStar;
}

void ViscSolver::PrintBanner()
{
    std::cout << "\n\n"
              << std::endl;
    std::cout << "# ---------------------------------------------------------------------------- #" << std::endl;
    std::cout << "#                                                                              #" << std::endl;
    std::cout << "#                                                   ///,        ////           #" << std::endl;
    std::cout << "#       Paul Dechamps                               \\  /,      /  >.           #" << std::endl;
    std::cout << "#       ULiege 2022                                  \\  /,   _/  /.            #" << std::endl;
    std::cout << "#       v0.1.0 - Memphis                              \\_  /_/   /.             #" << std::endl;
    std::cout << "#                                                      \\__/_   <               #" << std::endl;
    std::cout << "#                                                      /<<< \\_\\_               #" << std::endl;
    std::cout << "#   ██╗  ██╗ ██████╗ ██████╗ ██╗   ██╗███████╗        /,)^>>_._ \\              #" << std::endl;
    std::cout << "#   ██║  ██║██╔═══██╗██╔══██╗██║   ██║██╔════╝        (/   \\ /\\\\               #" << std::endl;
    std::cout << "#   ███████║██║   ██║██████╔╝██║   ██║███████╗            // ````              #" << std::endl;
    std::cout << "#   ██╔══██║██║   ██║██╔══██╗██║   ██║╚════██║     ======((`=======            #" << std::endl;
    std::cout << "#   ██║  ██║╚██████╔╝██║  ██║╚██████╔╝███████║     ___   ___   _               #" << std::endl;
    std::cout << "#   ╚═╝  ╚═╝ ╚═════╝ ╚═╝  ╚═╝ ╚═════╝ ╚══════╝    / __| | _ ) | |              #" << std::endl;
    std::cout << "#                                                | (__  | _ \\ | |__            #" << std::endl;
    std::cout << "#  High-Reynolds Operations for Robust Unsteady   \\___| |___/ |____|           #" << std::endl;
    std::cout << "#    and Separating Coupled Boundary layers                                    #" << std::endl;
    std::cout << "#                                                                              #" << std::endl;
    std::cout << "# ---------------------------------------------------------------------------- #" << std::endl;
    std::cout << "\n\n"
              << std::endl;
}