#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute lifting (linear or nonlinear) viscous flow around a NACA 0012
# Amaury Bilocq
#
# Test the viscous-inviscid interaction scheme
# Reference to the master's thesis: http://hdl.handle.net/2268/252195
# Reference test cases with Naca0012 (different from master's thesis):
# 1) Incompressible: Re = 1e7, M_inf = 0, alpha = 5°, msTE = 0.01, msLE = 0.01
#    -> nIt = 27, Cl = 0.55, Cd = 0.0058, xtrTop = 0.087, xtrBot = 0.741
#    -> msLe = 0.001, xtrTop = 0.061
# 2) Compressible: Re = 1e7, M_inf = 5, alpha = 5°, msTE = 0.01, msLE = 0.01
#    -> nIt = 33, Cl = 0.65, Cd = 0.0063, xtrTop = 0.057, xtrBot = 0.740
# 3) Separated: Re = 1e7, M_inf = 0, alpha = 12°, msTE = 0.01, msLE = 0.001
#    -> nIt = 43, Cl = 1.30 , Cd = 0.0108, xtrTop = 0.010, xtrBot = 1
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import dart.utils as floU
import dart.default as floD

import dart.pyVII.vUtils as viscU
import dart.pyVII.vAPI as viscAPI
import dart.pyVII.vCoupler as viscC
import dart.pyVII.vInterpolator as vInterpol

import numpy as np

import tbox
import tbox.utils as tboxU
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

import math

def main():
    print('Start')
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    Re = 1e7
    alpha = 10.*math.pi/180
    #alphaSeq = np.arange(-5, 5.5, 0.5)
    M_inf = 0.

    meshFactor = 1
    CFL0 = 1

    plotVar = 'H'

    # ========================================================================================

    c_ref = 1
    dim = 2
    
    if dim == 2:
        nSections = 1
    else:
        nSections = 10
        
    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'xLgt' : 5, 'yLgt' : 5, 'msF' : 1, 'msTe' : 0.01, 'msLe' : 0.01}
    msh, gmshWriter = floD.mesh(dim, 'models/n0012.geo', pars, ['field', 'airfoil', 'downstream'])
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, blw = floD.problem(msh, dim, alpha, 0., M_inf, c_ref, c_ref, 0.25, 0., 0., 'airfoil', te = 'te', vsc = True, dbc=True)
    tms['pre'].stop()

    # solve viscous problem

    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    
    vInterp = vInterpol.Interpolator(bnd)
    quit()
    vSolver = viscU.initBL(Re, M_inf, CFL0, nSections)
    iSolverAPI = viscAPI.dartAPI(floD.newton(pbl), blw[0], blw[1], nSections)
    coupler = viscC.Coupler(iSolverAPI, vSolver)

    #coupler.RunPolar(alphaSeq)
    coupler.Run()
    tms['solver'].stop()

    # extract Cp
    Cp = iSolverAPI.GetCp(bnd.groups[0].tag.elems)
    tboxU.write(Cp, 'Cp_airfoil.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('      Re        M    alpha       Cl       Cd      Cdp      Cdf       Cm')
    print('{0:6.1f}e6 {1:8.2f} {2:8.1f} {3:8.4f} {4:8.4f} {5:8.4f} {6:8.4f} {7:8.4f}'.format(Re/1e6, M_inf, alpha*180/math.pi, iSolverAPI.GetCl(), vSolver.Cdt, vSolver.Cdp, vSolver.Cdf, iSolverAPI.GetCm()))

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)
    print('SOLVERS statistics')
    print(coupler.tms)

    """# visualize solution and plot results
    #floD.initViewer(pbl)
    tboxU.plot(Cp[:,0], Cp[:,3], 'x', 'Cp', 'Cl = {0:.{3}f}, Cd = {1:.{3}f}, Cm = {2:.{3}f}'.format(iSolver.Cl, vSolver.Cdt, iSolver.Cm, 4), True)

    blScalars, blVectors = viscU.ExtractVars(vSolver)
    wData = viscU.Dictionary(blScalars, blVectors, xtr)
    viscU.PlotVars(wData, plotVar)"""

    
    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    if Re == 1e7 and M_inf == 0. and alpha == 2.*math.pi/180:
        tests.add(CTest('Cl', iSolverAPI.GetCl(), 0.2208, 5e-3))
        tests.add(CTest('Cd', vSolver.Cdt, 0.00531, 1e-3, forceabs=True))
        tests.add(CTest('Cdp', vSolver.Cdp, 0.0009, 1e-3, forceabs=True))
        tests.add(CTest('xtr_top', vSolver.Getxtr(0,0), 0.20, 0.03, forceabs=True))
        tests.add(CTest('xtr_bot', vSolver.Getxtr(0,1), 0.50, 0.03, forceabs=True))
    if Re == 1e7 and M_inf == 0. and alpha == 5.*math.pi/180:
        tests.add(CTest('Cl', iSolverAPI.GetCl(), 0.5488, 5e-3))
        tests.add(CTest('Cd', vSolver.Cdt, 0.0062, 1e-3, forceabs=True))
        tests.add(CTest('Cdp', vSolver.Cdp,0.0018, 1e-3, forceabs=True))
        tests.add(CTest('xtr_top', vSolver.Getxtr(0,0), 0.06, 0.03, forceabs=True))
        tests.add(CTest('xtr_bot', vSolver.Getxtr(0,1), 0.74, 0.03, forceabs=True))
    """else:
        raise Exception('Test not defined for this flow')"""

    tests.run()


    # eof
    print('')

if __name__ == "__main__":
    main()
